//
//  Network.m
//  Neuron1
//
//  Created by Leslie Smith on Mon Jun 07 2004.
//  Copyright (c) 2004 __MyCompanyName__. All rights reserved.
//

#import "Network.h"
#import "Neuron.h"
#import "Synapse.h"
#import "dynsynapse.h"
#import "shuntsynapse.h"
#import "Spikes.h"
#import "SpikeArray.h"
#import "NeuronView.h"




@implementation Network
- init: sender
{
	[super init] ;
	ViewObject = sender ;
	return self ;
}

- (void)dealloc { // not currently used, but should still clean up
	int index ;
	if (Synapses != nil) 
	{ // release all synapses then release the Synapses NSMutableArray (added 11 4 2008: might help memory leak)
		for (index = 0;index < [Synapses count]; index++)
			[[Synapses objectAtIndex: index] release] ;
		[Synapses release] ;
	}
	if (NeuronArray != nil) // release each neuron: might help memory leak
	{
		int index ;
		for (index=0; index < [NeuronArray count]; index++)
			[[NeuronArray objectAtIndex: index] release] ;
		[NeuronArray release] ;
	}
    [super dealloc];
}

- initSetRng: rngid
	/* set up random number generator class object */
{
	RandomNumberGenerator = rngid ;
	return self ;
}

- supplyParameters: (NSMutableDictionary *) SP: (NSMutableDictionary *) NeP: 
	(NSMutableDictionary *) NoP
	/* get the parameter Mutable Array */
{
	SimParameters = SP ;
	NeuronParameters = NeP ;
	NoiseParameters = NoP ;
	/*	[SimParameters retain] ;
	[NeuronParameters retain] ;
	[NoiseParameters retain] ;
	*/
	return self ;
}


- applySimParameters: (NSMutableDictionary *) SP 
{
	NSString *s1 ;
	//  updated May 28 2008
	SimParameters = SP ;
	int n_neuron_new ;
	// apply the simulation parameters
	s1 = [SimParameters valueForKey: [NSString stringWithCString: "Input Neurons"]] ;
	input_neuron = [s1 intValue] ;
	// how many neurons are there?
	s1 = [SimParameters valueForKey: [NSString stringWithCString: "Total Neurons"]] ;
	n_neuron_new = [s1 intValue] ;
	// line below modded LSS 14012008 so that there's no objection to changing the number
	// of neurons when the original number is 0 (which is the case at initialisation)
	if ((n_neuron != 0) && (n_neuron != n_neuron_new))
	{
		printf("%s", "NeuralNetwork:applySimParameters. Can't change number of neurons") ;
		[SimParameters setObject: [NSNumber numberWithInt: n_neuron] forKey: [NSString stringWithCString: "Total Neurons"]] ;
	}
	// how long should simulation last ?
	s1 = [SimParameters valueForKey: [NSString stringWithCString: "Duration"]] ;
	duration = [s1 floatValue] ;
	// what is the timestep
	s1 = [SimParameters valueForKey: [NSString stringWithCString: "Timestep"]] ;
	timestep = [s1 floatValue] ;
	// how many simulation steps
	simsteps = round(duration/timestep) ;
	return self ;
}

- applyNeuronParameters: (NSMutableDictionary *) NeP
{
	// apply the neuron parameters to the real (non-input) neurons
	NSString *s1 ;
	NSString *NonDefFilename ;
	FILE *nondefaultneurons ;
	
	// added 28 May 2008
	NeuronParameters = NeP ;
	int i ;
	float zerolevel, initphase, tonicvalue,threshold, refperiod, dissipation, squarediss ;
	float minact ;
	int neuronnumber ;
	neuronparams nptemp ;
	neuronparams NParams[[NeuronArray count]] ; // hold neuron parameters
	
	// read file of non default neurons
	NonDefFilename = [ViewObject getNonDefaultNeuronParamFile] ;
	
	s1 = [NeuronParameters valueForKey: [NSString stringWithCString: "Zero Level"]] ;
	zerolevel = [s1 floatValue] ;
	if (s1 == nil)  // added LSS 14012008
		NSLog(@"Network:applyNeuronParameters invoked before default neuron values set. Set default neuron parameters first") ;
	s1 = [NeuronParameters valueForKey: [NSString stringWithCString: "Initial Ph"]] ;
	initphase = [s1 floatValue] ;
	s1 = [NeuronParameters valueForKey: [NSString stringWithCString: "Tonic"]] ;
	tonicvalue = [s1 floatValue] ;
	s1 = [NeuronParameters valueForKey: [NSString stringWithCString: "Threshold"]] ;
	threshold = [s1 floatValue] ;
	s1 = [NeuronParameters valueForKey: [NSString stringWithCString: "Ref. Period"]] ;
	refperiod = [s1 floatValue] ;
	s1 = [NeuronParameters valueForKey: [NSString stringWithCString: "Dissipation"]] ;
	dissipation = [s1 floatValue] ;
	s1 = [NeuronParameters valueForKey: [NSString stringWithCString: "Sq. Diss'n"]] ;
	squarediss = [s1 floatValue] ;
	s1 = [NeuronParameters valueForKey: [NSString stringWithCString: "Min'm Activ'n"]] ;
	minact = [s1 floatValue] ;
	// apply these values to the neurons in the network
	for (i=input_neuron;i<[NeuronArray count]; i++)
	{ // set the default values into the array NParams
		NParams[i].zerolevel = zerolevel ;
		NParams[i].initphase = initphase ;
		NParams[i].tonicvalue = tonicvalue ;
		NParams[i].threshold = threshold ;
		NParams[i].refperiod = refperiod ;		
		NParams[i].dissipation = dissipation ;
		NParams[i].squarediss = squarediss ;
		NParams[i].minact = minact ;
	}	
	if (NonDefFilename != nil)
	{ // read in those which are different
	  // file format is <Neuron no> zerolevel initphase tonicvalue threshold refperiod dissipation squarediss minact
	  // as on the default cells
		nondefaultneurons = fopen([NonDefFilename UTF8String],"r") ; // open file
		while (fscanf(nondefaultneurons, "%d%f%f%f%f%f%f%f%f", &neuronnumber,
					  &nptemp.zerolevel,
					  &nptemp.initphase,
					  &nptemp.tonicvalue,
					  &nptemp.threshold,
					  &nptemp.refperiod,
					  &nptemp.dissipation,
					  &nptemp.squarediss,
					  &nptemp.minact) != EOF) // read neuron number
		{
			if ((neuronnumber < input_neuron) || (neuronnumber >= [NeuronArray count]))
				NSLog(@"Network.applyNeuronParameters: invalid neuron number %i \n", neuronnumber) ;
			else NParams[neuronnumber] = nptemp ;
		}
		fclose(nondefaultneurons) ;
	}
	for (i=input_neuron;i<[NeuronArray count]; i++)
	{
		[[NeuronArray objectAtIndex: i]  setRefPeriod: NParams[i].refperiod] ;	// before setting time step
		[[NeuronArray objectAtIndex: i]  setTimeStep: timestep] ; // actually from SimParameters: ALWAYS THE SAME
		[[NeuronArray objectAtIndex: i] setPhase: NParams[i].initphase] ;
		[[NeuronArray objectAtIndex: i] setTonic: NParams[i].tonicvalue] ;
		[[NeuronArray objectAtIndex: i] setThreshold: NParams[i].threshold] ;
		[[NeuronArray objectAtIndex: i] setRefPeriod: NParams[i].refperiod] ;
		[[NeuronArray objectAtIndex: i] setDissipation: NParams[i].dissipation] ;
		[[NeuronArray objectAtIndex: i] setSquareDissipation: NParams[i].squarediss] ;
		[[NeuronArray objectAtIndex: i] setMinimumActivation: NParams[i].minact] ;
		// noted 17 Jan 2008 that zerolevel was not used at all. Added it in
		[[NeuronArray objectAtIndex: i] setZerolevel: NParams[i].zerolevel] ;
	}
	//  temp
	// printf("%s%f%s","Network:applyNeuronParameters. dissipation = ", dissipation, "\n") ;
	return self ;
}

- applyNoiseParameters: (NSMutableDictionary *) NoP
{
	// apply the simulation parameters to the real (i.e. non-input) neurons
	NSString *s1 ;
	int i ;
	float tonicnoise, threshnoise, activnoise ;
	
	NoiseParameters = NoP ;
	
	s1 = [NoiseParameters valueForKey: [NSString stringWithCString: "Tonic N"]] ;
	tonicnoise = [s1 floatValue] ;
	s1 = [NoiseParameters valueForKey: [NSString stringWithCString: "Thresh N"]] ;
	threshnoise = [s1 floatValue] ;
	s1 = [NoiseParameters valueForKey: [NSString stringWithCString: "Activ N"]] ;
	activnoise = [s1 floatValue] ;
	// apply these values to the neurons in the network	
	if ((tonicnoise == 0) && (threshnoise == 0) && (activnoise == 0))
		for (i=input_neuron;i<[NeuronArray count]; i++)
			[[NeuronArray objectAtIndex: i] noNoise] ;
	else
		for (i=input_neuron;i<[NeuronArray count]; i++)	
		{
			[[NeuronArray objectAtIndex: i]  tonicNoise: tonicnoise] ;	
			[[NeuronArray objectAtIndex: i]  thresholdNoise: threshnoise] ;	
			[[NeuronArray objectAtIndex: i]  activationNoise: activnoise] ;	
		}
			
			return self ;
}

- (id) setup: (int) maxspikes : (int) keepAll: (NSMutableDictionary *) SP: (NSMutableDictionary *) NeP: (NSMutableDictionary *) NoP;

{
	NSString *s1 ;
	int i ;
	keepInputs = keepAll ;
	
	SimParameters = SP ;
	NeuronParameters = NeP ;
	NoiseParameters = NoP ;
	
	maxtotalspikes = maxspikes ;
	if (SimParameters == nil)
	{
		printf("%s", "Neuralnetwork: setup. SimParameters is nil. \n") ;
		return self ;
	}
	
	
	// set up output spike array
	if (spikesout != nil) free(spikesout) ;
	spikesout = (spikearray) malloc(sizeof(spiketype) * maxspikes) ;
	
	// how many input neurons are there?
	s1 = [SimParameters valueForKey: [NSString stringWithCString: "Input Neurons"]] ;
	input_neuron = [s1 intValue] ;
	// how many neurons are there?
	s1 = [SimParameters valueForKey: [NSString stringWithCString: "Total Neurons"]] ;
	n_neuron = [s1 intValue] ;
	// how long should simulation last ?
	s1 = [SimParameters valueForKey: [NSString stringWithCString: "Duration"]] ;
	duration = [s1 floatValue] ;
	// what is the timestep
	s1 = [SimParameters valueForKey: [NSString stringWithCString: "Timestep"]] ;
	timestep = [s1 floatValue] ;
	// how many simulation steps
	simsteps = round(duration/timestep) ;
	
	// set up the network
	// the 1st input_neuron are input neurons, and the following
	// n_neuron - input_neuron are normal neurons
	if (NeuronArray != nil) // release each neuron: might help memory leak
	{
		int index ;
		for (index=0; index < [NeuronArray count]; index++)
			[[NeuronArray objectAtIndex: index] release] ;
		[NeuronArray release] ;
	}
	NeuronArray = [NSMutableArray arrayWithCapacity: n_neuron] ;
	[NeuronArray retain] ;
	for (i=0;i<input_neuron; i++)
		[NeuronArray addObject: [[InputNeuron new] init]] ;
	for (i=input_neuron;i<n_neuron;i++)
		[NeuronArray addObject: [[[Neuron new] init] initSetRng: RandomNumberGenerator]] ;
	
	[self applyNeuronParameters: NeuronParameters] ;
	[self applyNoiseParameters: NoiseParameters] ;
	
	
	
	
	if (s1 == nil)
		printf("%s","Network:setup  s1 is nil\n") ;
	else
		printf("%s%i%s", "Network: simsteps = ",simsteps, "\n") ;
	
	// [self displayDictionary: SimParameters] ;
	return self ;
}

- (int) start
{
	// modified 7 9 2004: added local pool so that machine doesn't run out of memory so much
	int nspikes ; nspikes = 0 ;
	int noutspikes; noutspikes = 0 ;
	int i,j ;
	double actime ;
	NSAutoreleasePool *SubPool ;
	
	// run the simulation for simsteps time steps.
	
	Neuron *ThisNeuron;
	//input spike
	int nextinputspike ;
	nextinputspike = 0 ; // point to start of input spikes (may be null)
	int spikesnow[input_neuron] ; // to hold input spikes this last timestep
								  //output spikes
	spikearray nextspike ; 
	int neuron_number ;
	
	int n_outerloop, n_innerloop, whatsleft, loopno, innerno; // used in counting simulation steps
	
	// idea is to set up & release SubPool every N_SUBPOOL simulation steps
	// this means a double loop for the simulation
	n_outerloop = simsteps/N_SUBPOOL ; // integer division
	whatsleft = simsteps % N_SUBPOOL ;
	if (whatsleft > 0) n_outerloop++ ;
	
	nextspike = spikesout ; // point to start of output spike array
							// for ( i=0;i<simsteps;i=i+1) originally
	for (loopno=0;loopno<n_outerloop;loopno++)
	{
		if ((loopno == (n_outerloop - 1)) && (whatsleft > 0))
			n_innerloop = whatsleft ;
		else
			n_innerloop = N_SUBPOOL ;
		// allocate subpool
		SubPool = [[NSAutoreleasePool alloc] init] ;
		// now inner loop
		for (innerno=0;innerno<n_innerloop; innerno++)
		{
			i = (loopno * N_SUBPOOL) + innerno ; // i is simulation step count
			neuron_number = 0 ;
			// need the actual time
			actime = i * timestep ;
			// find out what, if any external inputs have arrived in last timestep
			for (j=0;j<input_neuron;j++) spikesnow[j] = 0 ; //set all to 0 first
			if (nextinputspike < number_input_spikes) // at end of input spike array
			{
				// fill spikesnow array with spikes that have occurred this timestep
				while ( (nextinputspike < number_input_spikes) && (actime >= inspikes[nextinputspike].timeofspike))
				{
					spikesnow[inspikes[nextinputspike].neuron] = 1 ; // just 1, however many there are
					nextinputspike++ ;
				}
				
			}
			// run through neurons				
			NSEnumerator *enumerator = [NeuronArray objectEnumerator];
			while (ThisNeuron = [enumerator nextObject]) {
				// note: if ThisNeuron is an input neuron we need to supply external input
				// do input spikes separately from normal ones
				if ([ThisNeuron isInputNeuron])
				{   
					if ([ThisNeuron runStep: spikesnow[neuron_number] : actime: i] == 1)
					{
						if (keepInputs) // we don't always record the input spikes
										// record spike times
						{ // records input spikes as well: will be optional
							nspikes++ ;
							nextspike->neuron = neuron_number;
							nextspike->timeofspike = i * timestep;
							nextspike++ ;
							if (nspikes == maxtotalspikes)
								printf("%s%i%s","Network.start: Too many spikes", nspikes, 
									   ". Increase MAXSPIKES in NeuronView.h. \n") ; 
							
						}	
					}
					
				}
				else // normal neuron
					if ([ThisNeuron runStep: 0: actime: i] == 1) // record spike times
					{ // records non-input spikes 
						nspikes++ ;
						noutspikes++ ;
						nextspike->neuron = neuron_number;
						nextspike->timeofspike = i * timestep;
						nextspike++ ;
						if (nspikes == maxtotalspikes)
							printf("%s%i%s","Network.start: Too many spikes", nspikes, 
								   ". Increase MAXSPIKES in NeuronView.h. \n") ; 
						
					}				
				// printf("%s%i%s", " Network:start timestep=", i, "\n") ;
				// printf("%s%i%s", " Network:start Neuron=", neuron_number, "\n") ;
				neuron_number++ ;
			}
			// enumerator = [NeuronArray objectEnumerator]; // back to start of array
			// printf("%s%i%s","runstep = ", result, "\n") ;
		}
		[SubPool release] ;
	}
	
	NSLog(@"Network:start. spike  total = %i output spikes = %i\n", nspikes, noutspikes) ;
	return nspikes ;
}

- (spikearray) returnSpikes 
{
	return spikesout ;
}

- setSynapses: (NSString *) synapsefile
	// read the synapse file (currently an NSString)
{
	FILE *sfile; // synapse file: text file
	syndatatype syn1 ;
	synapse * synapse1 ;
	dynsynapse *synapse2 ;
	shuntsynapse *synapse3;
	synapse * ThisSynapse ;
	Neuron * ThisNeuron ;
	BOOL adaptive, resetonnextinput;
	int dsteps, index ;
	int maxsynapses ;
	
	// read the synapses into the Synapses NSArray object
	sfile = fopen([synapsefile UTF8String],"r") ;
	if (Synapses != nil) 
	{ // release all synapses then release the Synapses NSMutableArray (added 11 4 2008: might help memory leak)
		for (index = 0;index < [Synapses count]; index++)
			[[Synapses objectAtIndex: index] release] ;
		[Synapses release] ;
	}
	// Synapses = [NSMutableArray arrayWithCapacity: 0.25 *( n_neuron*n_neuron)] ; // 25% of full rank as guide
	// better estimate: use factor 1 * (n_neuron - input_neuron)^ 2 + factor 2 * (n_neuron - input_neuron) * input_neuron
	maxsynapses = (FACTOR_INPUTNEURON * (n_neuron - input_neuron) * (n_neuron - input_neuron)) +
		(FACTOR_NONINPUTNEURON * (n_neuron - input_neuron) * input_neuron) ;
	Synapses = [NSMutableArray arrayWithCapacity: 	maxsynapses] ;
	// modified to include synapsetype 2 8 2004 LSS
	strcpy(syn1.synapsetype, "    ") ; // initialise
	while (fscanf(sfile, "%s%d%d%f%e%e", syn1.synapsetype,  &syn1.presyn, &syn1.postsyn,  &syn1.weight, 
				  &syn1.alpha, &syn1.delay) != EOF)
	{
		// check: input neurons are not allowed to be post-synaptic
		if (syn1.postsyn < input_neuron)
			printf("%s%s%d%s%d%s\n", "Network:setSynapses: input neurons may not be post-synaptic ",
				   "synapse from ", syn1.presyn , " to ", syn1.postsyn, "ignored") ;
		else
		{				
			dsteps = (int)ceil(syn1.delay/timestep) ; // get delay in timesteps
													  // should really do the following using objects...but I'm still an old C hacker.
													  // is this synapse adaptive?
			if ((syn1.synapsetype[1] == 'a') || (syn1.synapsetype[2] == 'a') || (syn1.synapsetype[3] == 'a'))
				adaptive  = YES ;
			else adaptive  = NO ;
			//does this synapse reset previous alpha-functions on the next input?
			if ((syn1.synapsetype[1] == 'r') || (syn1.synapsetype[2] == 'r') || (syn1.synapsetype[3] == 'r'))
				resetonnextinput  = YES ;
			else resetonnextinput  = NO ;
			// set up synapse, then store  in NSMutableArray Synapses
			if (syn1.synapsetype[0] == 's')
			{   // plain synapse: type = 1
				synapse1 = [[synapse new] init];
				[synapse1 setWeight: syn1.weight] ;
				[synapse1 setSynapseType: 1] ; // plain synapse
											   // part below removed
											   // make alpha function reset to 0 on firing for excitatory synapses, but not for inhibitory ones
											   //if (syn1.weight >= 0)
											   //	[synapse1 setAlpha: syn1.alpha alphareset: YES] ;
											   // else
				[synapse1 setAlpha: syn1.alpha alphareset: NO] ;
				[synapse1 computeSynapseEffects: timestep] ; // 
				if (resetonnextinput) 
					[synapse1 setResetContributionOnInputSpike: YES] ;
				else [synapse1 setResetContributionOnInputSpike: NO] ;
				[synapse1 setPresynNeuron: [NeuronArray objectAtIndex: syn1.presyn] number: syn1.presyn] ;
				[synapse1 setPostsynNeuron: [NeuronArray objectAtIndex: syn1.postsyn] number: syn1.postsyn ];
				// set delay
				[synapse1 setDelay: syn1.delay : dsteps ] ; // nil for now
				[synapse1 setAdaptive: adaptive : adaptvalues] ;
				[Synapses  addObject: synapse1] ;
			}
			else if (syn1.synapsetype[0] == 'd')
			{   // depressing synapse: type = 2
				synapse2 = [[dynsynapse new] init];
				// fill up rest as before
				[synapse2 setWeight: syn1.weight] ;
				[synapse2 setSynapseType: 2] ; // depressing synapse
											   // part below removed
											   // make alpha function reset to 0 on firing for excitatory synapses, but not for inhibitory ones
											   // if (syn1.weight >= 0)
											   // 	[synapse2 setAlpha: syn1.alpha alphareset: YES] ;
											   // else
				[synapse2 setAlpha: syn1.alpha alphareset: NO] ;
				[synapse2 computeSynapseEffects: timestep] ; // 
				if (resetonnextinput) 
					[synapse2 setResetContributionOnInputSpike: YES] ;
				else [synapse2 setResetContributionOnInputSpike: NO] ;
				[synapse2 setPresynNeuron: [NeuronArray objectAtIndex: syn1.presyn] number: syn1.presyn] ;
				[synapse2 setPostsynNeuron: [NeuronArray objectAtIndex: syn1.postsyn] number: syn1.postsyn];
				[synapse2 setDelay: syn1.delay : dsteps] ; // nil for now
				[synapse2 setAdaptive: adaptive : adaptvalues] ;
				// then read rest of synapse data line
				if (fscanf(sfile, "%f%f", &syn1.rechargerate, &syn1.utilrate) == 2)
				{
					[synapse2 setRechargerate: syn1.rechargerate] ;
					[synapse2 setUtilrate: syn1.utilrate] ;
					[Synapses  addObject: synapse2] ;
				}
				else printf("%s%i%s%i\n", "Network:setsynapses. Error in reading all of depressing synapse information: synapse from",
							syn1.presyn, " to " , syn1.postsyn) ;
			}
			else if (syn1.synapsetype[0] == 'm')
			{// shunting synapse
				synapse3 = [[shuntsynapse new] init];
				// fill up rest as before
				[synapse3 setWeight: syn1.weight] ;
				[synapse3 setSynapseType: 3] ; // shunting synapse
											   // make alpha function reset to 0 on firing for excitatory synapses, but not for inhibitory ones
											   // and shunting is always inhibitory
				[synapse3 setAlpha: syn1.alpha alphareset: NO] ;
				[synapse3 computeSynapseEffects: timestep] ; // 
				if (resetonnextinput) 
					[synapse3 setResetContributionOnInputSpike: YES] ;
				else [synapse3 setResetContributionOnInputSpike: NO] ;
				[synapse3 setPresynNeuron: [NeuronArray objectAtIndex: syn1.presyn] number: syn1.presyn] ;
				[synapse3 setPostsynNeuron: [NeuronArray objectAtIndex: syn1.postsyn] number: syn1.postsyn];
				[synapse3 setDelay: syn1.delay : dsteps] ; // nil for now
				[synapse3 setAdaptive: adaptive: adaptvalues] ;
				[Synapses  addObject: synapse3] ;
				
			} // if
		} // if
		strcpy(syn1.synapsetype, "    ") ; // re-initialise
	} // while
	fclose(sfile) ; // close the synapse file
	[Synapses retain] ;
	// remove all synapses first
	NSEnumerator *enumneuron = [NeuronArray objectEnumerator] ;
	while (ThisNeuron = [enumneuron nextObject])
		[ThisNeuron removeSynapsesAndTargets] ;
	// set up the synapses inside the neurons: go through the Synapses structure and insert pointers in Neurons
	NSEnumerator *enumerator = [Synapses objectEnumerator] ;
	while (ThisSynapse = [enumerator nextObject])
	{ // note that the presynaptic neuron is the one providing spikes to the neuron
	  // and  therefore the synapse is its target
		[[ThisSynapse getPresynNeuron] addTarget: ThisSynapse] ; // link synapse to presyn  neuron
		[[ThisSynapse getPostsynNeuron] addSynapse: ThisSynapse] ; // link synapse to postsynaptic neuron
	}
	return self ;
}

- setAdaptiveParams: (STDPinfo) adaptparams
{
	adaptvalues = adaptparams ;
	return self ;
}

- displaySynapses: (NSTextView *)SynapseDisplay 
{
	// permit display of synapse info in  its own window.
	syndatatype syn1 ;
	char tempcstring[100] ;
	synapse *ThisSynapse ;
	NSEnumerator *enumerator = [Synapses objectEnumerator] ;
	NSString *OneSynapse ;
	NSString *Separator ;
	NSRange endRange;
	while (ThisSynapse = [enumerator nextObject])
	{ 
		syn1 = [ThisSynapse getSynapseData] ;
		if (syn1.synapsetype[0] != 'd')
			sprintf(tempcstring, "%s\t%i\t%i\t%f\t%f\t%f\n",
					syn1.synapsetype, syn1.presyn, syn1.postsyn, syn1.weight, syn1.alpha, syn1.delay) ;
		else
			sprintf(tempcstring, "%s\t%i\t%i\t%f\t%f\t%f\t%f\t%f\n",
					syn1.synapsetype, syn1.presyn, syn1.postsyn, syn1.weight, syn1.alpha, syn1.delay,
					syn1.rechargerate, syn1.utilrate) ;
		// and save this information to the file
		OneSynapse = [NSString stringWithCString:  tempcstring];
		
		endRange.location = [[SynapseDisplay textStorage] length];
		endRange.length = 0;
		[SynapseDisplay replaceCharactersInRange:endRange withString:OneSynapse];
		endRange.length = [OneSynapse length];
	}
	//insert a seperator at the end of all the synapse values
	Separator = [NSString stringWithCString:  "---\n"];
	endRange.location = [[SynapseDisplay textStorage] length];
	endRange.length = 0;
	[SynapseDisplay replaceCharactersInRange:endRange withString:Separator];
	endRange.length = [Separator length];
	[SynapseDisplay scrollRangeToVisible:endRange]; 
	return self ;
}

- saveSynapses: (NSString *) savespikefile 
	// save the synapses into a text file
{
	FILE *outsfile;
	syndatatype syn1 ;
	synapse *ThisSynapse ;
	NSEnumerator *enumerator = [Synapses objectEnumerator] ;
	// open file for writing
	outsfile = fopen([savespikefile UTF8String],"w") ;
	while (ThisSynapse = [enumerator nextObject])
	{ 
		syn1 = [ThisSynapse getSynapseData] ;
		// and save this information to the file
		fprintf(outsfile, "%s\t%i\t%i\t%f\t%f\t%f\t",
				syn1.synapsetype, syn1.presyn, syn1.postsyn, syn1.weight, syn1.alpha, syn1.delay) ;
		if (syn1.synapsetype[0] == 'd')
			fprintf(outsfile,"%f\t%f", syn1.rechargerate, syn1.utilrate) ;
		
		fprintf(outsfile,"\n") ;		
	}
	fclose(outsfile) ;
	return self ;
}

- setInputSpikesFromArchive: (NSString *)spikefilename
{
	Spikes *InputSpikes ;
	SpikeArray *InputSpikeArray ;
	spikearray spiketemp ;
	int n_input_input_neurons, n_total_input_neurons,i, j, adjust ;
	// read the file
	InputSpikes = [NSKeyedUnarchiver unarchiveObjectWithFile: spikefilename] ;
	InputSpikeArray = [InputSpikes returnSpikeArrayObject] ;
	// empty dictionary
	if (InputSpikeInformation != nil)
		[InputSpikeInformation removeAllObjects] ;
	number_input_spikes = [InputSpikeArray getNumspikes] ; // gets overwritten later
	InputSpikeInformation = [InputSpikes getGeneratingParams] ;
	n_input_input_neurons = [[InputSpikeInformation valueForKey: [NSString stringWithCString: "Input Neurons"]] 
		intValue];
	n_total_input_neurons = [[InputSpikeInformation valueForKey: [NSString stringWithCString: "Total Neurons"]]
		intValue] ;
	[InputSpikeInformation retain] ;
	inspikes = (spikearray)malloc(number_input_spikes * sizeof(spiketype)) ;
	spiketemp = [InputSpikeArray getInputSpikes] ;
	// copy spiketemp to inspikes adjusting neuron number
	// how to decide whether to adjust neuron number:  
	// if n_total_input_neurons == input_neuron don't adjust
	// if (n_total_input_neurons - n_input_input_neurons) == input_neuron subtract n_input_input_neurons
	if (n_total_input_neurons == input_neuron) adjust = 0 ;
	else if ((n_total_input_neurons - n_input_input_neurons) == input_neuron) adjust = n_input_input_neurons ;
	else 
	{
		NSLog(@"Network.setInputSpikesFromArchive input neurons numbers doesn't conform to no of input neurons") ;
		adjust = 0 ;
	}
	j=0 ;
	for (i=0;i<number_input_spikes;i++)
	{		
		if ((spiketemp[i].neuron - adjust) >= 0)
		{
			inspikes[j].neuron = spiketemp[i].neuron - adjust ;
			inspikes[j].timeofspike = spiketemp[i].timeofspike ;
			j = j + 1 ;
		}
	}
	number_input_spikes = j ;
	return self ;
}

- setSpikeInput: (NSString *)spikeinputfile 
	// supply input spikes
	// simple format: <neuronnumber> <time>
	//
{
	FILE *inspikefile ;
	char spike[] = "spikes" ;
	char pname[64] ;
	char pvalue[64] ;
	NSString *ParamName, *ParamValue ;
	int sno, number_of_spikes ;
	float timeofinputspike, lastspiketime ;
	int neuronnumber ;
	int inputoffset ; // to be subtracted from input neuron number
	float timemultiplier ; // times to be mulitplied by this (used when the times are in samples etc.)
	int ignoreabove ; // ignore neurons above this number
	
	lastspiketime = -1 ; // initialise
	inspikefile = fopen([spikeinputfile UTF8String],"r") ;
	if (inspikefile == NULL)
	{
		printf("%s %s\n", "Network.setSpikeInput: unable to open spike input file (plain) ", [spikeinputfile UTF8String]) ;
		return self ;
	}
	// file now open
	// how many spikes (maximum) are there?
	fscanf(inspikefile, "%s", pname) ;
	// printf("%s\n", "Network.setInputSpikes point 3") ;
	// create or empty dictionary
	if (InputSpikeInformation == nil)
		InputSpikeInformation =  [NSMutableDictionary dictionaryWithCapacity: 30] ;
	else [InputSpikeInformation removeAllObjects] ;
	
	while (strcmp(pname, spike) != 0)
	{ // store the parameter
		ParamName = [NSString stringWithCString: pname] ;
		if (fscanf(inspikefile, "%s", pvalue) == EOF)
		{
			printf("%s\n", "Network.setSpikeInput: input spike file ended before spikes found") ;
			[InputSpikeInformation dealloc] ;
			return self;
		}
		ParamValue = [NSString stringWithCString: pvalue] ;
		[InputSpikeInformation setObject:ParamValue forKey:ParamName] ;
		// temp
		// printf("%s %s %s\n", "Network.setInputSpikes. pname, pvalue", pname, pvalue) ;
		if (fscanf(inspikefile, "%s", pname) == EOF)
		{
			printf("%s\n", "Network.setSpikeInput: input spike file ended before spikes found") ;
			[InputSpikeInformation dealloc] ;
			return self ;
		}
	}
	number_of_spikes = [[InputSpikeInformation valueForKey: [NSString stringWithCString: "nspikes"]] intValue] ;
	inputoffset = 0 ;
	// if skipover parameter is set read it
	if ([InputSpikeInformation objectForKey: @"skipover"])
		inputoffset = [[InputSpikeInformation valueForKey: @"skipover"] intValue] ;
	// if timemultiplier value is set read it
	timemultiplier = 1 ;
	if ([InputSpikeInformation objectForKey: @"timemultiplier"])
		timemultiplier = [[InputSpikeInformation valueForKey: @"timemultiplier"] floatValue] ;
	// allow to ignore neurons above a certain number
	ignoreabove = 1000000 ; // need to alter if > 1000000 neurons!
	if ([InputSpikeInformation objectForKey: @"ignoreabove"])
		ignoreabove = [[InputSpikeInformation valueForKey: @"ignoreabove"] intValue] ;	
	// allocate space for the spikes
	// ugh...malloc...
	if (inspikes != nil) free(inspikes) ; // deallocate
	sno = 0 ;
	inspikes = (spikearray)malloc(number_of_spikes * sizeof(spiketype)) ;
	while (( sno < number_of_spikes) && (fscanf(inspikefile, "%d%f\n", &neuronnumber, &timeofinputspike) != EOF))
	{
		// check that this spike time is after or same as last spike time
		// debug
		// printf("%s\t%i\t%f\n", "input ", neuronnumber, timeofinputspike) ;
		if (lastspiketime > timeofinputspike)
			printf("%s\t%i\t%f\n", "Network.setSpikeInput: input spikes not in time order", neuronnumber, timeofinputspike) ;
		else
		if ((neuronnumber - inputoffset >= 0) && (neuronnumber <= ignoreabove))
		{ // store this spike time
			inspikes[sno].neuron = neuronnumber - inputoffset ;
			inspikes[sno].timeofspike = timeofinputspike * timemultiplier ;
			lastspiketime = inspikes[sno].timeofspike ;
			sno = sno + 1 ;
		}
	}
	fclose(inspikefile) ;
	// temp
	// [self displayDictionary: InputSpikeInformation] ;
	// sno is number of spikes actually read, and number_of_spikes is number promised
	// should be the same, but might not be. Best to use the smaller of the two
	if (number_of_spikes > sno) number_input_spikes = sno ;
	else number_input_spikes = number_of_spikes ; // set instance variable
	[InputSpikeInformation retain] ;
	return self ;
}



- setInputSpikes: (NSString *)inputspikefile 
	// read in the input spikes, ready to apply them to input neurons
	// NB: specialised for the output spike file from generateonsetspikes1_mono
	// input neurons are ordered 0..number_of_bands-1 (highest sensitivity level), 
	// number_of_bands..(number_of_bands*2)-1 (next s-level),
	// ...
	// number_of_bands*(number_of_levels-1)..(number_of_bands*number_of_levels)-1 (lowest s-level)
	//
{
	FILE *inspikefile ;
	char pname[64] ;
	char pvalue[64] ;
	NSString *ParamName, *ParamValue ;
	char spike[] = "spikes" ;
	float lastspiketime ;
	int level, band ;
	float timeofinputspike ;
	int sno, number_of_spikes, number_of_levels, number_of_bands ;
	
	// temp
	// printf("%s\n", "Network.setInputSpikes") ;
	lastspiketime = -1 ;
	// read all the parameters and store in another NSDictionary
	// create or empty dictionary
	if (InputSpikeInformation == nil)
		InputSpikeInformation =  [NSMutableDictionary dictionaryWithCapacity: 30] ;
	else [InputSpikeInformation removeAllObjects] ;
	// open the file for reading
	// printf("%s\n", "Network.setInputSpikes point 1") ;
	inspikefile = fopen([inputspikefile UTF8String],"r") ;
	// printf("%s %s\n", "Network.setInputSpikes point 2", [inputspikefile cString]) ;
	if (inspikefile == NULL)
	{
		printf("%s %s\n", "Network.setInputSpikes: unable to open spike input file (MatLab) ", [inputspikefile UTF8String]) ;
		return self ;
	}
	// printf("%s %i\n", "Network.setInputSpikes point 2a", (int) inspikefile) ;
	fscanf(inspikefile, "%s", pname) ;
	// printf("%s\n", "Network.setInputSpikes point 3") ;
	while (strcmp(pname, spike) != 0)
	{ // store the parameter
		ParamName = [NSString stringWithCString: pname] ;
		if (fscanf(inspikefile, "%s", pvalue) == EOF)
		{
			printf("%s\n", "Network.setInputSpikes: input spike file ended before spikes found") ;
			return self;
		}
		ParamValue = [NSString stringWithCString: pvalue] ;
		[InputSpikeInformation setObject:ParamValue forKey:ParamName] ;
		// temp
		// printf("%s %s %s\n", "Network.setInputSpikes. pname, pvalue", pname, pvalue) ;
		if (fscanf(inspikefile, "%s", pname) == EOF)
		{
			printf("%s\n", "Network.setInputSpikes: input spike file ended before spikes found") ;
			return self ;
		}
	}
	// read the spikes themselves
	// also check that the number of input neurons seems correct
	number_of_spikes = [[InputSpikeInformation valueForKey: [NSString stringWithCString: "nspikes"]] intValue] ;
	number_of_levels = [[InputSpikeInformation valueForKey: [NSString stringWithCString: "iterations"]] intValue] ;
	number_of_bands = [[InputSpikeInformation valueForKey: [NSString stringWithCString: "channels"]] intValue] ;
	if (input_neuron != (number_of_levels * number_of_bands))
		printf("%s%i\n", "Network.setInputSpikes. # of bands * # of levels not same as # input units",
			   input_neuron) ;
	// allocate space for the spikes
	// ugh...malloc...
	if (inspikes != nil) free(inspikes) ; // deallocate
	
	inspikes = (spikearray)malloc(number_of_spikes * sizeof(spiketype)) ;
	sno = 0 ;
	while (( sno < number_of_spikes) && (fscanf(inspikefile, "%d%d%f", &level, &band, &timeofinputspike) != EOF))
	{
		// check that this spike time is after or same as last spike time
		if (lastspiketime > timeofinputspike)
			printf("%s%f\n", "Network.setInputSpikes: input spikes not in time order", timeofinputspike) ;
		else
		{// store this spike time
		 // levels and bands both start at 1
		 // choose one of the two below:
			
			// order: level 1, bands 1-N; level 2, bands 1-N...
			// inspikes[sno].neuron = ((level-1)*number_of_bands) + (band - 1) ;
			// order: band 1,levels 1-M, band 2, levels 1-M,...
			inspikes[sno].neuron = ((band-1)*number_of_levels) + (level - 1) ;
			inspikes[sno].timeofspike = timeofinputspike ;
			lastspiketime = timeofinputspike ;
			sno = sno + 1 ;
		}
	}
	fclose(inspikefile) ;
	// temp
	// [self displayDictionary: InputSpikeInformation] ;
	// sno is number of spikes actually read, and number_of_spikes is number promised
	// should be the same, but might not be. Best to use the smaller of the two
	if (number_of_spikes > sno) number_input_spikes = sno ;
	else number_input_spikes = number_of_spikes ; // set instance variable
	[InputSpikeInformation retain] ;
	return self ;
}

- displayDictionary: Dict1
	// display a dictionary
{
	int nelements ;
	NSEnumerator *enumerator ;
	nelements = [Dict1 count] ;
	printf("%s%i%s", "Network: displayDictionary no of elements = ", nelements, "\n") ;
	enumerator = [Dict1 keyEnumerator];
	id key, valueforkey ;
	
	while ((key = [enumerator nextObject])) {
		/* code that uses the returned key */ 
		valueforkey = [Dict1 valueForKey: key] ;
		printf("%s%s", "displayDictionary key = ", [key  cString] ) ;
		printf("%s%s%s", " value for key = ", 
			   [valueforkey  cString] , "\n") ;
	}
	return self;
}

- (int) getNumberOfInputNeurons 
{
	return input_neuron ;
}

- (int) getNumberOfNeurons 
{
	return n_neuron ;
}



@end
