
#import "Neuron.h"
#import <math.h>
#import "Random.h"
#import "synapse.h"
#import "dynsynapse.h"

@implementation Neuron

- init
	/* initialise on setup */
{
	[super init] ;
	[self initial: DEFTHRESH: DEFDISS: DEFMINACT: DEFREFPD] ; 
	tonicnoise = 0 ;
	threshnoise = 0 ;
	PreSynaptic = [NSMutableArray arrayWithCapacity: MAXPRESYNSYNAPSES] ;
//	PostSynaptic = [NSMutableArray arrayWithCapacity: 20] ; // in InputNeuron
	[PreSynaptic retain] ;
//	[PostSynaptic retain] ;
	return self ;
}

- (void)dealloc {
	[self removeSynapsesAndTargets] ;
	if (PreSynaptic != nil) [PreSynaptic release] ;
    [super dealloc]; // super is InputNeuron
}

- initial: (float) newthresh: (float) newdiss : 
           (float) minactivation: (float) refpd  /* initialise */
{
	threshold = newthresh ;
	dissipation = newdiss ;
	minact = minactivation ; 
	refractperiod = refpd ;
	activation = resetlevel ;
	timestep = DEFTIME ;
	output = 0 ;
	return self ;
}

- initSetRng: rngid
	/* set up random number generator class object */
{
	Rng = rngid ;
	return self ;
}

- (int) isInputNeuron 
{
	return NO ;
}

- removeSynapsesAndTargets 
// remove all the synapses and targhets prior to setting them all up
{
	if (PreSynaptic != nil) [PreSynaptic removeAllObjects] ;
	if (PostSynaptic != nil) [PostSynaptic removeAllObjects] ;
	return self ;
}

- addSynapse: Syn 
	// add a synapse to the set that this neuron owns (is postsynaptic to)
{
	[PreSynaptic addObject: Syn] ;
	return self ;
}

- addTarget: Syn  // add a synapse to the set that this neuron innervates (is presynaptic)
{
	[PostSynaptic addObject: Syn] ;
	return self ;
}

- setTimeStep: (float) stepsize  /* set timestep */
/* must be after setting of refractory period */
{
	timestep = stepsize ;
	if (refractperiod == 0) intrefperiod = 0 ;
	else intrefperiod = ceil(refractperiod/timestep) ;
	return self ;
}

- setRefPeriod: (float) refpd 
{
	refractperiod = refpd ;
	return self ;
}

- setTonic: (float) tonic /* tonic input */
{
	tonicinput = tonic ;
	return self ;
}

- setThreshold: (float) thresh 
{
	threshold = thresh ;
	return self ;
}

- setDissipation: (float) diss
{
	dissipation = diss ;
	return self ;
}

- setSquareDissipation: (float) squarediss 
{
	squaredissipation = squarediss ;
	return self ;
}

- setMinimumActivation: (float) minactivn 
{
	minact = minactivn ;
	return self ;
}

- setZerolevel: (float) zerolevel // added 17012008: set reset level for activation
{ 
	resetlevel = zerolevel ;
	return self ;
}

- setPhase: (float) phasepoint  /* set initial phase */
{
	/* since activation(t) = */
	/*                (tonicinput/dissipation)(1 - exp(- dissipation * t)) */
	/* and the period, P, can be calculated, we can compute */
	/* activation(t) = above, with period*phasepoint to be the activation */
	/* at this phase in the cycle */
	/* N.B. 0<= phasepoint < 1 */
	/* If tonic is 0, then don't use getPeriod! */
	/* instead, set activation to phase fraction of threshold (LSS 4.4.2004) */
	/* ignores square dissipation term */
	
	// updated 17 Jan2008: use simple linear phase. 
	activation = phasepoint *  (threshold - resetlevel) + resetlevel ;
	/* initphase = phasepoint ;
	if (tonicinput == 0) 
		activation = phasepoint *  (threshold - resetlevel) + resetlevel ;
	else
	{
		if (dissipation !=0)
			activation =  (tonicinput/dissipation) *
				(1 - exp(- dissipation * (phasepoint * [self getPeriod]))) ;
		else activation = tonicinput * (phasepoint * [self getPeriod]) ;
	}
	*/
	return self ;
}

- resetNeuron
{
	output = 0 ; 
	return self ;
}

- noNoise
{
	nonoise = YES ;
	return self ;
}

- tonicNoise: (float) level 
	/* set up fractional noise on tonic input */
{
	nonoise = NO ;
	tonicnoise = level ;
	return self ;
}

- thresholdNoise: (float) level 
	/* set up noise on threshold */
{
	nonoise = NO ;
	threshnoise = level ;
	return self ;
}

- activationNoise: (float) level 
	/* set up noise on threshold */
{
	nonoise = NO ;
	activationnoise = level ;
	return self ;
}

- (float) getActivation  /* returns activation level */
{
	return activation ;
}

- (float) getRefPeriod
{
	return refractperiod ;
}

- (int) getOutput   /* returns output */
{
	return output ;
}

- (float) getTimeStep 
{
	return timestep ;
}

- (float) getDissipation
{
	return dissipation ;
}

- (float) getSquareDissipation
{
	return squaredissipation ;
}

- (float) getThreshold 
{ 
	return threshold ;
}

- (float) getTonic
{
	return tonicinput ;
}

- (float) getPeriod
	/* calculate and return period */
	/* ignores square term */
{
	float period ;
	/* assumes constant tonic input and nothing else! */
	if (dissipation == 0)
		period = threshold/tonicinput ;
	else
		period = (1.0 / dissipation) *
			log(tonicinput/(tonicinput - 
							(dissipation * threshold))) ;
	/* add refractory period */
	period += refractperiod ;
	return period ;
}

- (float) getPhase
{
	return initphase ;
}

/* run for 1 timestep */
/* returns output value */
- (int) runStep:   (int) otherinput : (double) actime: (int) stepno
// revised version: LSS 17 6 2004
{
	float tonic, thrshld, syninput, shuntvalue ;
	if (refractcount != 0)
	{
		/* in refractory period */
		// though we ignore synaptic input, still calculate it to avoid multiple presynaptic events during RP
		syninput = [self calcSynContribution: actime: stepno] ;
		shuntvalue = [self calcMultSynContribution: actime : stepno] ;
		refractcount-- ;
		output = 0;
		activation = resetlevel ; // altered 17012008 LSS
		/* check minimum activation not exceeded */
		if (activation < minact) activation = minact ;
		return 0 ;
	}
	
  // calculate synaptic contribution
	syninput = [self calcSynContribution: actime: stepno] ;
	shuntvalue = [self calcMultSynContribution: actime : stepno] ;
	// temp
	// if (shuntvalue != 1)
	// 		NSLog(@"Neuron:runStep. Shuntvalue = %f at %i \n", shuntvalue,stepno) ;
	if (nonoise)
	{
	  // Note:  add all the inputs together
		activation = activation  + syninput + otherinput + // add synaptic input & tonic and external
		timestep * (tonicinput -  // tonic input is scaled by timestep
					(dissipation * activation) // as is linear dissipation
					- (squaredissipation * (activation * activation)) ) ; // and squared dissipation
		// apply shunting synapse
		activation  = activation * shuntvalue ; // should this be scaled by the time step?
		/* does it fire? */	
		if (activation >= threshold)
		{// fires
			[self fire: actime: stepno] ;
			/* set output */
			output = 1; 
			/* set refractory count */
			refractcount = intrefperiod ;
			/* reset activation to the inter-neuron input */
			if (refractcount == 0) activation = otherinput + resetlevel ;
			else activation = resetlevel ;
		}
		else output = 0 ;
		/* check activation  more than minimum activation  */
		if (activation < minact) activation = minact ;
		
	} // no noise
	else // there is noise
	{
		/* add tonic noise */
		if (tonicnoise != 0)
			tonic = (tonicinput * (1 - ([Rng plusminus] * tonicnoise)));
		else tonic = tonicinput ;
		
		/* add threshold noise */
		if (threshnoise != 0) 
			thrshld = (threshold * (1 + [Rng plusminus] * threshnoise));
		else thrshld = threshold ;	
		
		/* calc activation: add everything: (n.b. uses tonic, not original tonicinput */
		activation = activation  + syninput + otherinput + // add synaptic input & tonic and external
			timestep * (tonic -  // tonic input is scaled by timestep
						(dissipation * activation) // as is linear dissipation
						- (squaredissipation * (activation * activation)) ) ; // and squared dissipation
		
		/* add activation noise */
		if (activationnoise != 0)
			activation = (activation * (1 - ([Rng plusminus] * activationnoise))) ;
		// lastly apply shunting synapse
		activation  = activation * shuntvalue ;
		/* does it fire? */
		if (activation >= thrshld)
		{ // fires
			[self fire: actime: stepno] ;
			/* set output */
			output = 1; 
			/* set refractory count */
			refractcount = intrefperiod ;
			/* reset activation */
			if (refractcount == 0) activation = otherinput + resetlevel ;
			else activation = resetlevel ;
		}
		else  output = 0 ;
		
		/* check minimum activation not exceeded */
		if (activation < minact) activation = minact ;
	} // noise
	
// temp
// printf("%s%f%s","neuron:runstep. activation = ", activation, "\n") ;
	return output ;
}

- (float) calcSynContribution: (double) actime: (int) stepno
// calculate synaptic contribution by running through the presynaptic synapses
// and summing
// Note that this presumes additive synapses: multiplicative would need treated differently
{
	float synapticinput ;
	synapticinput = 0 ;
	id ThisSynapse ;
	// notify target synapses if any that this neuron has fired
	if (PreSynaptic != nil)
	{
		NSEnumerator *enumerator = [PreSynaptic objectEnumerator];
		while (ThisSynapse = [enumerator nextObject]) {
			// omit type 3 (shunting) synapses
			if ([ThisSynapse getSynapseType] != 3)
			synapticinput = synapticinput + [ThisSynapse getSynapticContribution: actime: stepno] ;
		}
	}
	return synapticinput ;
}

- (float) calcMultSynContribution: (double) actime: (int) stepno   // multiplicative synapses
// calc synaptic shuting effect of the multiplicatgive synapses
{
	float fracttoremain ;
	fracttoremain = 1 ;
	id ThisSynapse ;
	// notify target synapses if any that this neuron has fired
	if (PreSynaptic != nil)
	{
		NSEnumerator *enumerator = [PreSynaptic objectEnumerator];
		while (ThisSynapse = [enumerator nextObject]) {
			//use only type 3 (shunting) synapses
			if ([ThisSynapse getSynapseType] == 3)
				fracttoremain = fracttoremain * [ThisSynapse getSynapticContribution: actime: stepno] ;
		}
	}
	return fracttoremain ;
}


- fire : (double) actime : (int) stepno// action to take on firing at time actime
{
	id ThisSynapse ;
	// const char *temp ;
	// notify target synapses if any that this neuron has fired
	if (PostSynaptic != nil) // postsynaptic relative to this neuron
	{
		NSEnumerator *enumerator = [PostSynaptic objectEnumerator];
		while (ThisSynapse = [enumerator nextObject]) {
			// temporary
			// temp = [[ThisSynapse description] cString] ;
			// printf("%s%s\n", "Neuron:fire. postsynaptic synapse", temp) ;
			
			[ThisSynapse presynFires: actime: stepno] ; // presunaptic relative to the synapse
		}
	}
	// do we reset presynaptic synapses to 0, or do we let alpha-functions continue to contribute after firing?
	
	if (PreSynaptic != nil) // presynaptic relative to this neuron
	{
		NSEnumerator *enumerator = [PreSynaptic objectEnumerator];
		while (ThisSynapse = [enumerator nextObject]) {			
			[ThisSynapse postsynFires: actime: stepno] ; // postsynaptic relative to the synapse
		}
	}
	
	// for STDP, 
	// alter the method postsynFires in Synapse.m
	return self ;
}

@end
