/* NeuronView */
/* A first attempt at a simple neuron simulator */
/* Basic level system runs as of 6 April 2004: LSS */
/* Let's trry to save/load paramaters using NSMutableDictionary */
/* and NSString, NSNumber, NSSavePanel, NSLoadPanel */
/* started 7 April 2004 */

#import <Cocoa/Cocoa.h>
#import "Network.h"

#define MAXSPIKES 10000000
#define NPARAMS 20

@interface NeuronView : NSView
{
// id button ;
id DisplayRadio ;
Network *NeuralNetwork; /* whole neural network */
// panels
id InfoPanel ;
id STDPParametersPanel ;
id SynapseInfoPanel ;
id SynapseTextView ;
// parameter objects in InfoPanel
id NeuronParams ;
id NoiseSeeds ;
id NoiseParams ;
id SimParams ;
id DisplayTimes ; // start and stop times for display
id SpikesGenerated ;
// parameters objects in STDPParameters Panel
id PrePostNN ; // YES if Nearest Neighbour
id PostPreNN ;
id LTPWins ; // YES if LTP wins over LTD
id PrePostAdd ; // Yes for add, NO for multiply
id PostPreAdd ;
id TauValues ; // 0'th is pre->post, 1st is post->pre
id WeightValues ; // ditto
id WeightBoundaries ; // minimum and maximum values for adjustable weights
id ZxPermission ; // is zero crossing of weights permitted?
id ZeroAbsorbing ; // is zero an absorbing barrier?
id AntiHebbianInhibitory ; // are inhibitoy synapses to be adapted in an antiHebbian way?
id FileNameDisplay ; // for displaying spike and synapse file names
STDPinfo adaptparams ;

id Rng ; /* random number generator */
float simduration ;

int nspikes ; // number of spikes generated
int displayInputs ;
int inputsource ; // 0 not set, 1 MatLab input, 2 input from archived spikes

NSMutableDictionary *SimParamnames ;
NSMutableDictionary *NeuronParamnames ;
NSMutableDictionary *NoiseParamnames ;

int nparams ;
NSString *LastDirectory ; // last directory for parameters
NSString *LastSpikeOutputDirectory ;
NSString *LastInputSpikesDirectory ;
NSString *SynapseFile ; // filename containing synapse file
NSString *InputSpikeFile ; // filename for input spikes (MATLAB)
NSString *SpikeInputFile ; // filename for input spikes (plain)
NSString *ArchiveInputSpikes ; // filename for archived inout spikes
NSString *NonDefaultNeuronParamFile ; // filename for non-default neurons
NSRect image ; // image containing the spikes.

float startdisplaytime, enddisplaytime ;
float stepwidth ; // width in pixels/time

NSBezierPath *p1 ; // global bezier path for drawing
}


- (id) start: sender ;
- (id) reDisplay: sender ; // reinitialise network, but don't re-read files
- (id) setupNeuralNetwork: sender ;
- (id) setNewRngSeed: sender ;
- (id) showInfoPanel: sender ;
- (id) showSynapseInfoPanel: sender ;
- (id) clearSynapseInfo: sender ;
- (id) readNeuronParams: sender ;
- (id) saveNeuronParams: sender ;
- (id) setNonDefaultNeuronParamFile: sender ;
- (NSString *) getNonDefaultNeuronParamFile ;
- (id) loadNeuronParams: sender ;
- (id) showNeuronParameters: sender ;
- (id) readNoiseParams: sender ;
- (id) saveNoiseParams: sender ;
- (id) loadNoiseParams: sender ;
- (id) showNoiseParameters: sender ;
- (id) readSimParams: sender ;
- (id) saveSimParams: sender ;
- (id) loadSimParams: sender ;
- (id) showSimParameters: sender ;
- (id) displayDictionary: (NSDictionary *) dict1 ;
- (id) setSynapseFile: sender ;
- (id) saveSynapses: sender ;
- (NSString *) getSynapseFile ;
- (id) setInputSpikesFile: sender ; // Matlab file
- setSpikeInputFile: sender ; // plain input
- (NSString *) getInputSpikesFile ; // Matlab file
- (NSString *) getSpikeInputFile ; // plain input
- (id) saveSpikesGeneratedHumanReadable: sender  ; // in a readable form
- (id) saveSpikesGenerated: sender ;
- (id) loadSpikesGeneratedEarlier: sender ;
- (id) setDisplayTime: sender ;
- (id) savePDFDisplay: sender ;
- (void) fillBezierPath ;
- getSTDPParameters: sender ;

// delegate method
- (void)applicationDidFinishLaunching:(NSNotification *)aNotification ;

@end
