//
//  dynsynapse.m
//  Neuron1
//
//  Created by Leslie Smith on Sat Jul 31 2004.
//  Copyright (c) 2004 __MyCompanyName__. All rights reserved.
//

#import "dynsynapse.h"


@implementation dynsynapse

- init
	/* initialise on setup */
{
	[super init] ;
	level = 1 ; // start with undepressed synapse
	return self ;
}

- setRechargerate: (float) rate 
{
	rechargerate = rate ;
	return self ;
}

- setUtilrate: (float)  rate 
{
	utilrate = rate ;
	return self ;
}

- (float) getRechargerate ;
{ return rechargerate ;}

- (float) getUtilrate ;
{ return utilrate;}

- presynFires: (float) t_fire: (int) stepfire
	// notification that presynaptic neuron has fired
{
	int i ;	


	// calculate effective current level from level and previous spike time
	if (neverspiked)
		level = 1 ;
	else
	{
		// choose one of exponential or linear recharge (LSS 5 8 2004)
		//  linear chosen
		// exponential recharge
		// level = level * exp(rechargerate * (t_fire - lastspiketime)) ;
		// choose rechargerate so that exp(rechrgerate * (expected recovery interspike interval)) is
		// of the right size: 
		// but this is problematic: recharge time depends strongly on level that dep syn was reduced to.
		// if near 0, then recharge time is long
		// linear recharge
		level = level +  ((t_fire - lastspike) * rechargerate) ;
		// less problematic in above sense.
		// choose recharge rate so that (expected recovery interspike interval) * rechargerate is about 0.7
		if (level > 1) level = 1 ;
	}	
	
	[super presynFires : t_fire : stepfire] ;
	lastspike = t_fire ; // better than using lastspiketimes array as this one doesn't disappear.
	// shift up the effective weights
	for (i=presynspikes-1;i>0;i--)
	{
		w_effs[i] = w_effs[i-1] ;
	}
	// and insert the newest  effective weight then reduce level due to this spike discharging the reservoir
	w_effs[0] = weight * level ;
	level = level * exp(-utilrate) ;
	// if ((prenumber == 5) && (postnumber ==13))
	// NSLog(@"dynsynapse:presyn %i postsyn %i  Fires spike at %f weight = %f level = %f\n" , prenumber, 
	// 	  postnumber, t_fire, w_effs[0],level) ;
	return self ;
}


- (float) getSynapticContribution: (double) current_time: (int) stepno
{
	// updated to include delays (code almost identical to that in synapse.m)
	// only difference is w_weffs[i] instead of weight
	float contribution;
	int timesincespike ;
	int i,j, deduct ;
	if ((neverspiked) || (presynspikes == 0))
	{
		contribution = 0 ;
		return 0 ;
	}
	if (do_reset_alpha) 
	{
		// remove those presynaptic inputs which have arrived
		// but not those in the delay (i.e. in presynaptic axon)
		for (i=presynspikes-1;i>=0;i--)
			if ((stepno - lastspikesteps[i]) > 0)
				presynspikes-- ;
		contribution = 0 ;	
		do_reset_alpha = NO ; // turn off alpha resetting here.
	}
	else
	{
		// add up contribution of all presynaptic spikes
		contribution = 0 ;
		for (i=0; i< presynspikes; i++)
		{
			timesincespike = (stepno - lastspikesteps[i]) ;
			if (isadaptive && (timesincespike == 0)) // may do STDP
				[self applySTDPPresynaptically : lastspiketimes[i] effectiveweight: w_effs[i]];
			if ((timesincespike >= 0) && (timesincespike < lengthalphaeffect)) // ignore if negative: in delay
				contribution = contribution + w_effs[i] * alphaarray[stepno - lastspikesteps[i]] ;
		}	
			
		// delete presyn spikes that have come to an end
		// there may be > 1	
		while ((presynspikes > 0) && ((stepno - lastspikesteps[presynspikes - 1]) >= lengthalphaeffect))
		{			
			presynspikes = presynspikes - 1 ;
			// update history only if adaptive
			if (isadaptive)
			{
				lastusedspiketime = lastspiketimes[presynspikes] ;
				lastusedspikestep = lastspikesteps[presynspikes] ; // stores most recent last spike
				lastusedspikew_eff =  w_effs[presynspikes] ;
				// keep a longer history too, used for pre->post classic STDP 
				// slide history up, deleting any which are too old
				deduct = 0 ;
				for (j=lastusednumber;j>0;j--)
				{
					// only keep them if they are within tikme relevant to pre->post classic STDP
					// bug fix 22012008 (MAXPRESYNDT * adaptvalues.prepostTau) replaces (MAXPRESYNDT / adaptvalues.prepostTau)
					if ( (current_time - lastusedspiketimearray[j-1]) < (MIN(0.1, MAXPRESYNDT / adaptvalues.prepostTau)))
					{
						lastusedspiketimearray[j] = lastusedspiketimearray[j-1] ;
						lastusedspikesteparray[j] = lastusedspikesteparray[j-1] ;
						lastusedspikew_effarray[j] = lastusedspikew_effarray[j-1] ;
					}
					else deduct++ ;
				}
				lastusednumber = lastusednumber - deduct ;
				lastusedspiketimearray[0] = lastusedspiketime ;
				lastusedspikesteparray[0] = lastusedspikestep ;
				lastusedspikew_effarray[0] = lastusedspikew_eff ;
				if (lastusednumber == MAXKEPTSPIKES - 1)
					NSLog(@"DynSynapse:getSynapticContribution run out of space in lastusedspiketimearray. %i\n",
						  lastusednumber) ;
				else
				lastusednumber++ ;
			}
		}
	}
	
	return contribution;
}

- applySTDPPresynaptically : (float) presyntime effectiveweight: (float) effweight 
// compute and apply post=>pre STDP LTD: for depressing synapses
// effweight is the effective weight for this synapse
{	float scale ;
	float oldweight = weight ;
	
	
	if (lastpostsyntime < 0) return self ;
	if (adaptvalues.LTPwinner && postsynusedadaptively) return self ;
	if ((lastpostsyntime - presyntime) > (MAXPOSTSYNDT / adaptvalues.postpreTau))
		return self ;
	if (adaptvalues.postpreNN && postsynusedpresynaptically) return self ;
		// if this synapse is excitatory, and zero-crossing is not permitted, and the weight is 0, leave it there.
	if (adaptvalues.excitatory && !adaptvalues.zxpermitted && (weight == 0)) return self ;
	// if this one is inhibitory, and zero-crossing is not permitted, and this one is antihebbian (so LTP)
	// and the weight is 0, leave it there.
	if (!adaptvalues.excitatory && !adaptvalues.zxpermitted && 
		adaptvalues.antihebbianforinhibitory && (weight == 0)) return self ;	
	// go ahead and alter weight	
	// set scale up for when it's needed.
	if ((weight == 0) || (effweight == 0))
		scale = 1 ;
	else scale = effweight/weight ; 
	
	if (adaptvalues.postpreadditive) // weight change scaled by w_eff/weight
		weight = weight - adaptvalues.postpreDeltaWeight * 
			exp(- fabs((lastpostsyntime - presyntime)) * adaptvalues.postpreTau) * scale ;
	else // multiplicative
		 // note that weight must be < 1
		 // weight change multiplier is w_eff not weight
		weight = weight - (adaptvalues.postpreDeltaWeight * 
						   exp(- fabs((lastpostsyntime - presyntime)) * adaptvalues.postpreTau) * weight * effweight) ;
	
	if (!adaptvalues.zxpermitted) // weight may not change sign
	{
		if (((oldweight < 0) && (weight >= 0)) ||
			((oldweight > 0) && (weight <= 0 ))) // NB can be going either way here
		{
			weight = 0 ;
			if (adaptvalues.zeroabsorbing)
			[self setAdaptive: NO : adaptvalues] ; // absorbing boundary
		}
	}	
	// check weight not above (antihebbian)( boundary
	if (weight > adaptvalues.maxallowedweight) weight = adaptvalues.maxallowedweight ;	
	// and not below either 
	if (weight < adaptvalues.minallowedweight) weight = adaptvalues.minallowedweight ;			
	postsynusedpresynaptically = YES ;
				
	return self ;
}

- postsynFires: (float) t_fire: (int) stepfire
	// Use for resetting alpha, and (later) for STDP
{
	float dt ;
	int i, ii ;
	float alphamult ;
	float oldweight = weight ;
	int validpresynspikes ;
	float dtsforadaptation[MAXCONCURRENTPRESYN+MAXKEPTSPIKES] ; // absolute possible max size
																// record time of last spike
	float w_effsforadaptation[MAXCONCURRENTPRESYN+MAXKEPTSPIKES] ;
	float effweight ; // effective weight for use in STDP
	float scale ; // will be used to hold effweight/weight, or 1 if either are 0
					  // record time of last spike
	lastpostsyntime = t_fire ;
	lastpostsynstep = stepfire ;
	postsynusedpresynaptically = NO ;
	postsynusedadaptively = NO ;
	dt = 0 ; // it will get overwritten
	// default is reset excitatory synapses, but not inhibitory ones
	if (alpha_reset) do_reset_alpha = YES  ; // ignore previous spikes
	if (isadaptive)
	{
		// adaptivity goes here
		// if this synapse is inhibitory, and zero-crossing is not permitted, and the weight is 0, leave it there.
		if (!adaptvalues.excitatory && !adaptvalues.zxpermitted && (weight == 0)) return self ;
		// if this one is excitatory, and zero-crossing is not permitted, and this one is antihebbian (so LTD)
		// and the weight is 0, leave it there.
		if (adaptvalues.excitatory && !adaptvalues.zxpermitted && 
			adaptvalues.antihebbianforinhibitory && (weight == 0)) return self ;		
		// if there are spikes to be processed, and the most recent one has actually arrived,
		// or if there were spikes some time ago
		// set scale up for when it's needed.
		if (presynspikes == 0) effweight = weight ;
			else effweight = w_effs[presynspikes - 1] ;
		if ((weight == 0) || (effweight == 0))
			scale = 1 ;
		else scale = effweight/weight ; 
		if (((presynspikes > 0) && ((t_fire - lastspiketimes[presynspikes-1]) > 0)) || (lastusedspiketime > 0)) 
			// < 0 implies no presynaptic spikes
		{
			if (adaptvalues.prepostNN)
			{
				// find the time since last spike, and relevant last effective weight
				if ((presynspikes > 0) && ((t_fire - lastspiketimes[presynspikes-1]) > 0))
					
				{ // check: there may be a number of spikes currently inside the alpha function
					ii = presynspikes-1 ;
					dt = (t_fire - lastspiketimes[ii]) ; // keep compiler happy
					while ((ii >= 0) && ((t_fire - lastspiketimes[ii]) > 0))
					{
						dt = (t_fire - lastspiketimes[ii]) ;
						effweight = w_effs[ii] ;
						ii = ii - 1 ;
					}
				}
				
				else 
				{
					dt = (t_fire - lastusedspiketime)  ;
					effweight = lastusedspikew_eff ;
				}
				if (dt < MAXPRESYNDT / adaptvalues.prepostTau) // do nothing if time since last presynaptic spike was too long
				{
					if (adaptvalues.prepostadditive) // scale weight change 
					{
						if (alpha != 0)
							weight = weight + adaptvalues.prepostDeltaWeight * 
								exp(- fabs(dt - (1/alpha)) * adaptvalues.prepostTau) * scale;
						else 
							weight = weight + adaptvalues.prepostDeltaWeight * 
								exp(- fabs(dt) * adaptvalues.prepostTau) * scale;
					}
					else // multiplicative
						 // note that weight must be < 1
					{
						if (alpha != 0)
						weight = weight + (adaptvalues.prepostDeltaWeight * 
										   exp(- fabs(dt - (1/alpha)) * adaptvalues.prepostTau) * 
										   (1-weight) * effweight) ;
						else
							weight = weight + (adaptvalues.prepostDeltaWeight * 
											   exp(- fabs(dt) * adaptvalues.prepostTau) * 
											   (1-weight) * effweight) ;
					}
					postsynusedadaptively = YES ;
					if (!adaptvalues.zxpermitted) // weight may not change sign
					{
						if (((oldweight < 0) && (weight >= 0)) ||
							((oldweight > 0) && (weight <= 0 ))) // NB can be going either way here
						{
							weight = 0 ;
							if (adaptvalues.zeroabsorbing)
							[self setAdaptive: NO : adaptvalues] ; // absorbing boundary
						}
					}	
					// check weight not above boundary
					if (weight > adaptvalues.maxallowedweight) weight = adaptvalues.maxallowedweight ;
					// and not below either (antihebbian could cause this)
					if (weight < adaptvalues.minallowedweight) weight = adaptvalues.minallowedweight ;					
				}
			}
			else  // classic
			{
				// needs to consider both spikes still within their alpha function (in lastspiketimes)
				// and spikes whose alpha function has finished (in lastusedspiketimearray)
				// find the time since last spikes
				validpresynspikes = 0 ; // use to count the presaynaptic spikes helping this Dw
				if ((presynspikes > 0) && ((t_fire - lastspiketimes[presynspikes-1]) > 0))
				{ // check: there may be a number of spikes currently inside the alpha function
					ii = presynspikes-1 ;
					dt = (t_fire - lastspiketimes[ii]) ; // keep compiler happy
					while ((ii >= 0) && ((t_fire - lastspiketimes[ii]) > 0))
					{
						dt = (t_fire - lastspiketimes[ii]) ;
						dtsforadaptation[validpresynspikes] = dt ;
						w_effsforadaptation[validpresynspikes] = w_effs[ii] ;
						validpresynspikes++ ;
						ii = ii - 1 ;
					}
				}
				// now find those whose alpha function has ended
				for (i=0;i<lastusednumber;i++)
				{
					dtsforadaptation[validpresynspikes] = (t_fire - lastusedspiketimearray[i]) ;
					w_effsforadaptation[validpresynspikes] = lastusedspikew_effarray[i] ;
					validpresynspikes++ ;
				}
				// now do the actual weight update
				for (i=0;i<validpresynspikes;i++)
				{
					if (dt < MAXPRESYNDT / adaptvalues.prepostTau)
					{	
						// bug fix 20 Jan 2005
						if ((weight == 0) || (w_effsforadaptation[i] == 0))
							scale = 1 ;						
						else scale = w_effsforadaptation[i]/weight ;

						// bug fix 23 Jan 2005
						if (alpha == 0) alphamult = 0 ;
						else
							alphamult = (1/alpha) ;
						
						if (adaptvalues.prepostadditive)
							weight = weight + adaptvalues.prepostDeltaWeight * 
								exp(- fabs(dtsforadaptation[i] - alphamult) * adaptvalues.prepostTau) *
								scale ;
						else // multiplicative
							 // note that weight must be < 1
							weight = weight + (adaptvalues.prepostDeltaWeight * 
											   exp(- fabs(dtsforadaptation[i] - alphamult) * adaptvalues.prepostTau) * 
											   (1-weight) * w_effsforadaptation[i]) ;
						postsynusedadaptively = YES ;

					}
				}
				if (!adaptvalues.zxpermitted) // weight may not change sign
				{
					if (((oldweight < 0) && (weight >= 0)) ||
						((oldweight > 0) && (weight <= 0 ))) // NB can be going either way here
					{
						weight = 0 ;
						if (adaptvalues.zeroabsorbing)
						[self setAdaptive: NO : adaptvalues] ; // absorbing boundary
					}
				}	
				// check weight not above boundary
				if (weight > adaptvalues.maxallowedweight) weight = adaptvalues.maxallowedweight ;	
				// and not below either (antihebbian could cause this)
				if (weight < adaptvalues.minallowedweight) weight = adaptvalues.minallowedweight ;	
			} // classic
		}
	}
	return self ;
}

- (syndatatype) getSynapseData 
{	syndatatype syn1 ;
	
	syn1 = [super getSynapseData] ;

	// fix those that need fixed in syn1 
	syn1.synapsetype[0] = 'd';	
	// add in those that are new
	syn1.rechargerate = rechargerate ;
	syn1.utilrate = utilrate ;
	return syn1 ;
}

@end
